;; Generated package description from firefox-javascript-repl.el  -*- no-byte-compile: t -*-
(define-package "firefox-javascript-repl" "0.9.5.0.20230605.161924" "Jack into Firefox" '((emacs "26.1")) :commit "fee99c20f1f4684ed845eeac9eda46fcde6bf1e9" :url "https://elpa.gnu.org/packages/firefox-javascript-repl.html" :authors '(("Thomas Fitzsimmons" . "fitzsim@fitzsim.org")) :maintainer '("Thomas Fitzsimmons" . "fitzsim@fitzsim.org"))
